#!/usr/bin/env python

# Copyright 2011 Ryan 'The Mad Crapper' Whelan. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are
# permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice, this list of
#      conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright notice, this list
#      of conditions and the following disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY RYAN WHELAN ''AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL RYAN WHELAN OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are those of the
# authors and should not be interpreted as representing official policies, either expressed
# or implied, of Ryan Whelan.

import re
import os
from subprocess import Popen, PIPE

checkfreq = 300.0
svnbin = '/usr/bin/svn'
localdir = '/home/hlplayer/LF/orangebox/luminousforts'
gamedir = '/home/hlplayer/LF/orangebox'
gamecmd = 'screen -D -m '+gamedir+'/srcds_run -game luminousforts -console -allowdebug +map lfc_InnerComplex +maxplayers 16 -autoupdate -tickrate 67'

### Vague Exceptions ###

class RepoException(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return repr(self.msg)

class RemoteRepoException(Exception):
    def __init__(self):
        raise RepoException("Error Parsing Remote Repo")

class LocalRepoException(Exception):
    def __init__(self):
        raise RepoException("Error Parsing Local Repo")


class ServerNotRunning(Exception):
    def __str__(self):
        return repr('Server Not Running')

### Screwy Class Usage ###

class SVN(object):
    def __init__(self, local, svnbin):
        self.svnbin = svnbin
        self.localrepo = local

    def _findRev(self, repo):
        cmdoutput = Popen([self.svnbin, 'info', repo],
                          stdout = PIPE).communicate()[0]

        if not cmdoutput: return False

        rev = re.findall('Revision: ([0-9]*)\n', cmdoutput)[0]
        if not rev: return False

        return int(rev)

    def _findURL(self):
        cmdoutput = Popen([self.svnbin, 'info', self.localrepo],
                          stdout = PIPE).communicate()[0]

        if not cmdoutput: return False

        url = re.findall('URL: (.*)\n', cmdoutput)[0]
        if not url: raise LocalRepoException
        return url


    def needupdate(self):
        localrev = self._findRev(self.localrepo)
        remoterev = self._findRev(self._findURL())
        if not localrev:
            raise LocalRepoException
            return False

        if not remoterev:
            raise RemoteRepoException
            return False
        
        if remoterev > localrev: return True
        return False

    def update(self):
        cmdoutput = Popen([self.svnbin, 'up', self.localrepo],
                  stdout = PIPE).communicate()[0]

        return self._findRev(self.localrepo)


class GameServer(object):
    def __init__(self, gamedir, gamecmd):
        self.gamecmd = gamecmd
        self.running = False
        os.chdir(gamedir)

    def run(self):
        # No error handling so as to pass an exception
        # back to the caller. Hopefully print to the terminal
        # before we fork()
        self.gameproc = Popen(gamecmd.split(), stdin = open(os.devnull, 'r'),
                                               stderr = open(os.devnull, 'w'))
        self.running = True
        return True

    def stop(self):
        if not self.running: raise ServerNotRunning
        self.gameproc.terminate()

        # I don't like it either
        while True:
            try:
                if os.waitpid(-1,os.WNOHANG) == (0, 0):
                    self.gameproc.kill()
                    continue
            except:
                self.running = False
                return True


if __name__ == '__main__':
    import time
    import sys
    svn = SVN(localdir, svnbin)
    gameserver = GameServer(gamedir, gamecmd)

    if svn.needupdate(): svn.update()

    if os.fork(): os._exit(0)

    # Dog Speed; good luck
    os.closerange(0, 255)
    os.setsid()

    # Who needs Logging anyway?.....
    sys.stdin = open(os.devnull, 'r')
    sys.stdout = open(os.devnull, 'w')
    sys.stderr = open(os.devnull, 'w')
    # ... i kind of do. Will do later....

    if os.fork(): os._exit(0)

    # Logging really needs to be done.
    # We don't know if the game server startup fails
    gameserver.run()

    while True:
        time.sleep(checkfreq)
        if svn.needupdate():
            # We should send a message to the server
            # saying we are exiting for update...
            gameserver.stop()
            svn.update()
            gameserver.run()
        else:
            continue


